library walletpackage;

import 'dart:math';
import 'dart:typed_data';

import 'package:bitcoin_flutter/bitcoin_flutter.dart' as bitcoin_flutter;
import 'package:bip39/bip39.dart' as bip39;
import 'package:web3dart/crypto.dart' as web3dart_crypto;
import 'package:web3dart/web3dart.dart' as web3dart;
import 'package:rlp/rlp.dart' as rlp;

class WalletData {
  String mnemonic;//助记词，本地不保存
  String keystore;
  Map<String, String> addresses = new Map<String, String>();
}

class CoinInfo {
  final String path;
  final bitcoin_flutter.NetworkType network;
  final int decimals;
  final String address;
  final int gasLimit;

  CoinInfo(this.path, this.network, this.decimals,
      {this.address, this.gasLimit});
}

final litcoin = new bitcoin_flutter.NetworkType(
    messagePrefix: '\x19Litecoin Signed Message:\n',
    bech32: 'ltc',
    bip32:
    new bitcoin_flutter.Bip32Type(public: 0x019da462, private: 0x019d9cfe),
    pubKeyHash: 0x30,
    scriptHash: 0x32,
    wif: 0xb0);

final bitcoincash = bitcoin_flutter.NetworkType(
    messagePrefix: '\x18Bitcoin Signed Message:\n',
    bech32: 'bc',
    bip32:
    new bitcoin_flutter.Bip32Type(public: 0x0488b21e, private: 0x0488ade4),
    pubKeyHash: 0x00,
    scriptHash: 0x05,
    wif: 0x80);

final Map<String, CoinInfo> _BitcoinType = {
  'BTC': CoinInfo("m/44'/0'/0'/0/0", bitcoin_flutter.bitcoin, 8),
  'LTC': CoinInfo("m/44'/2'/0'/0/0", litcoin, 8),
  'BCH': CoinInfo("m/44'/145'/0'/0/0", bitcoincash, 8),
};
final Map<String, CoinInfo> _OmniType = {
  'USDT':
  CoinInfo("m/44'/0'/0'/0/1", bitcoin_flutter.bitcoin, 8, address: '31'),
};
final Map<String, CoinInfo> _EthereumType = {
  'ETH': CoinInfo("m/44'/60'/0'/0/0", null, 18, gasLimit: 21000),
  'ETC': CoinInfo("m/44'/61'/0'/0/0", null, 18, gasLimit: 21000),
};
final Map<String, CoinInfo> _Erc20Type = {
  'EUSDT': CoinInfo("m/44'/60'/0'/0/1", null, 6,
      address: '0xdAC17F958D2ee523a2206206994597C13D831ec7', gasLimit: 60000),
};

class Wallet {
  static WalletData createMnemonicAndAddress(String passwd) {
    String mnemonic = bip39.generateMnemonic();
    return importMnemonicAndAddress(mnemonic, passwd: passwd);
  }

  static bitcoin_flutter.HDWallet _getChildNode(
      bitcoin_flutter.HDWallet hdWallet, String type, Map<String, CoinInfo> map) {
    if (map.containsKey(type)) {
      CoinInfo coinInfo = map[type];
      hdWallet.network = coinInfo.network;
      bitcoin_flutter.HDWallet childnode = hdWallet.derivePath(coinInfo.path);
      return childnode;
    } else {
      return null;
    }
  }

  static web3dart.EthereumAddress _node2EtherumAddress(
      bitcoin_flutter.HDWallet ethnode) {
    if (ethnode == null) return null;
    web3dart.EthereumAddress ethaddr = web3dart.EthereumAddress(web3dart_crypto
        .publicKeyToAddress(web3dart_crypto.privateKeyBytesToPublic(
        web3dart_crypto.hexToBytes(ethnode.privKey))));
    return ethaddr;
  }

  static void _addBitcoinAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type, Map<String, CoinInfo> map) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type, map);
    wd.addresses.addAll({type: node.address});
  }

  static void _addEtherumAddress(
      WalletData wd, bitcoin_flutter.HDWallet hdWallet, String type, Map<String, CoinInfo> map) {
    bitcoin_flutter.HDWallet node = _getChildNode(hdWallet, type, map);
    wd.addresses.addAll({type: _node2EtherumAddress(node).hexEip55});
  }

  static WalletData importMnemonicAndAddress(String mnemonic, {String passwd = ''}) {
    if (bip39.validateMnemonic(mnemonic)) {
      WalletData wd = new WalletData();
      wd.mnemonic = mnemonic;

      bitcoin_flutter.HDWallet hdWallet =
      bitcoin_flutter.HDWallet.fromSeed(bip39.mnemonicToSeed(wd.mnemonic));

      web3dart.Wallet wallet = web3dart.Wallet.createNew(
          web3dart.EthPrivateKey.fromHex(hdWallet.seed),
          passwd,
          Random.secure());
      wd.keystore = wallet.toJson();
      print(wd.keystore);

      _BitcoinType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type, _BitcoinType);
      });
      _OmniType.keys.forEach((type) {
        _addBitcoinAddress(wd, hdWallet, type, _OmniType);
      });
      _EthereumType.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type, _EthereumType);
      });
      _Erc20Type.keys.forEach((type) {
        _addEtherumAddress(wd, hdWallet, type, _Erc20Type);
      });
      print(wd.addresses);

      return wd;
    } else {
      return null;
    }
  }

  static String changePassword(
      String keystore, String passwd, String newPasswd) {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      web3dart.Wallet newWallet = web3dart.Wallet.createNew(
          wallet.privateKey, newPasswd, Random.secure());
      return newWallet.toJson();
    } catch (e) {
      print('signTransaction error : $e');
    }
    return null;
  }

  static Future<String> _signBitcoinTransaction(
      bitcoin_flutter.HDWallet wallet, String txhex) async {
    try {
      bitcoin_flutter.ECPair ecPair = bitcoin_flutter.ECPair.fromPrivateKey(
          web3dart_crypto.hexToBytes(wallet.privKey));
      bitcoin_flutter.Transaction transaction =
      bitcoin_flutter.Transaction.fromHex(txhex);
      bitcoin_flutter.TransactionBuilder transactionBuilder =
      bitcoin_flutter.TransactionBuilder.fromTransaction(transaction);
      transactionBuilder.inputs.forEach((element) {
        int index = transactionBuilder.inputs.indexOf(element);
        transactionBuilder.sign(vin: index, keyPair: ecPair);
      });
      String hex = transaction.toHex();
      return hex;
    } catch (e) {
      print('_signBitcoinTransaction error : $e');
    }
    return null;
  }

  static Future<String> _signEthereumTransaction(
      web3dart.EthPrivateKey privateKey, String txhex) async {
    var res = await privateKey.sign(web3dart_crypto.hexToBytes(txhex));
    print(res);
    return null;
  }

  static Future<String> signTransaction(
      String type, String keystore, String passwd, String txhex) async {
    try {
      web3dart.Wallet wallet = web3dart.Wallet.fromJson(keystore, passwd);
      bitcoin_flutter.HDWallet hdWallet =
      bitcoin_flutter.HDWallet.fromSeed(wallet.privateKey.privateKey);
      bitcoin_flutter.HDWallet childWallet;
      String signedTxhex = null;
      if (childWallet != null) {
        if (_BitcoinType.keys.contains(type) || _OmniType.keys.contains(type)) {
          print(childWallet.address);
          signedTxhex = await _signBitcoinTransaction(childWallet, txhex);
        } else if (_EthereumType.keys.contains(type) || _Erc20Type.keys.contains(type)) {
          print(_node2EtherumAddress(childWallet).hexEip55);
          web3dart.EthPrivateKey privateKey =
          web3dart.EthPrivateKey.fromHex(childWallet.privKey);
          signedTxhex = await _signEthereumTransaction(privateKey, txhex);
        }
      }
      return signedTxhex;
    } catch (e) {
      print('signTransaction error : $e');
    }
    return null;
  }

  static String _createBitcoinTransaction(List ins, List outs, int fee) {
    return null;
  }

  static String createEthereumTransaction(
      String from, String to, int value, int gasPrice) {
    return null;
  }

  static String createOmniTransaction(
      List ins, List outs, int fee, int value, int contract) {
    return null;
  }


  static String _stripZeros(String value) {
    var result = (value);
    if (result.length == 0) {
      return result;
    }
    // Find the first non-zero entry
    var start = 0;
    while (result[start] == 0) {
      start++;
    }
    // If we started with zeros, strip them
    if (start > 0) {
      result = result.substring(start);
    }
    return result;
  }
  static String _hexlify(dynamic value) {
    print('-------------------------------');
    print(value);
    if (value is int) {
      value = BigInt.from(value);
    }
    String hex = '';
    if(value is BigInt){
      hex = value.toRadixString(16);
    }
    if (value is Uint8List) {
      hex = web3dart_crypto.bytesToHex(value);
    }
    if (value is String) {
      var tmp = web3dart_crypto.hexToBytes(value);
      hex = web3dart_crypto.bytesToHex(tmp);
    }
    print(hex);
    if (hex.length % 2 == 1) {
      hex = '0' + hex;
    }
    hex = '0x' + hex;
    print(hex);
    print('-------------------------------');
    return hex;
  }
  static _arrayify(tmp){
    return tmp;
  }
  static void _addRawlist(List raw, dynamic field, int length, int maxLength){
    var value = _arrayify(_hexlify(field));
    // Fixed-width field
    if (length > 0 && value.length != length && value.length > 0) {
      print('invalid length 1');
    }
    // Variable-width (with a maximum)
    if (maxLength > 0) {
      value = _stripZeros(value);
      if (value.length > maxLength) {
        print('invalid length 2');
      }
    }
    raw.add(_hexlify(value));
  }
  static String createErc20Transaction(
      String type, String from, String to, int value, int gasPrice, int nonce) {
    if (_Erc20Type.keys.contains(type)) {
      CoinInfo coinInfo = _Erc20Type[type];

      String operationHex =
          '0xa9059cbb000000000000000000000000' + to.toLowerCase().substring(2);
      String valueHex = value.toRadixString(16);
      String tokenHex =
      ('0000000000000000000000000000000000000000000000000000000000000000' +
          valueHex)
          .substring(valueHex.length);
      String dataHex = operationHex + tokenHex;

//      web3dart.Transaction tx = new web3dart.Transaction(
//          from: web3dart.EthereumAddress.fromHex(from),
//          to: web3dart.EthereumAddress.fromHex(coinInfo.address),
//          maxGas: coinInfo.gasLimit,
//          gasPrice:
//              web3dart.EtherAmount.inWei(BigInt.from(gasPrice * 1000000000)),
//          value: web3dart.EtherAmount.zero(),
//          data: web3dart_crypto.hexToBytes(dataHex),
//          nonce: nonce);


      print(to);
      var chars = to.substring(2).split('');
      var hashed = Uint8List(40);
      for (var i = 0; i < 40; i++) {
        hashed[i] = chars[i].codeUnitAt(0);
      }
      hashed = _arrayify(web3dart_crypto.keccak256(hashed));
      for (var i = 0; i < 40; i += 2) {
        if ((hashed[i >> 1] >> 4) >= 8) {
          chars[i] = chars[i].toUpperCase();
        }
        if ((hashed[i >> 1] & 0x0f) >= 8) {
          chars[i + 1] = chars[i + 1].toUpperCase();
        }
      }
      to = '0x' + chars.join('');
      print(to);
      print(to.length);
      return null;

      List raw = new List();
      _addRawlist(raw, nonce, 0, 32);
      _addRawlist(raw, BigInt.from(gasPrice * 1000000000), 0, 32);
      _addRawlist(raw, coinInfo.gasLimit, 0, 32);
      _addRawlist(raw, to, 20, 0);
      _addRawlist(raw, 0, 0, 32);
      _addRawlist(raw, dataHex, 0, 0);
      var unsignedTransaction = web3dart_crypto.bytesToHex(rlp.Rlp.encode(raw));

      return unsignedTransaction;

    } else {
      return null;
    }
  }

}
